import re
import os
import functools
import threading
from queue import Queue
from colour_printing.style import setting
from colour_printing.custom.config import Config
import time

INFO = 'INFO'
DEBUG = 'DEBUG'
SUCCESS = 'SUCCESS'
ERROR = 'ERROR'
WARN = 'WARN'
tip = "[*]Tip>> {}"


class PrintMeError(Exception):
    def __init__(self, message):
        super().__init__(message)


def level_wrap(func):
    @functools.wraps(func)
    def wrap(self, *args, **kwargs):
        if self.switch is False or self.Master_switch is False:
            return
        if func.__name__ in self.filter:
            return
        return func(self, *args, **kwargs)

    return wrap


def log_to(printme, log_name: str):
    count = 5
    path = os.path.join(printme.root_path, log_name)
    print(f'[*]Tip>> 日志文件path: {path}')
    with open(path, 'a+') as f:
        while printme.switch and printme.Master_switch:
            if count < 0:
                print(end='\r')
                print('[*]Tip>> 日志输出关闭')
                break
            if printme.queue.empty():
                print(end='\r')
                print(f'[*]Tip>> 等待输出信息 {count} s', end='')
                time.sleep(1)
                count -= 1
                continue
            f.write(printme.queue.get())
            f.flush()


class PrintMe(object):
    level_list = [INFO, ERROR, SUCCESS, DEBUG, WARN]
    Master_switch = True

    def __init__(self, template: str, config_filename: str = 'colour_printing_config.py', log_output: bool = False,
                 log_name: str = 'colour_printing_log.log'):
        self.raw_template = template
        self.term = re.findall(r'(?<=\{)[^}]*(?=\})+', template)
        if "message" not in self.term:
            raise PrintMeError('\n [*]Tip>> template muse have {message} ! ')
        term_wrap = {i: "{%s}{%s}{%s}" % (i + '0', i, i + '1') for i in self.term}
        self.template = template.format(**term_wrap)
        # store
        self.box = {}
        self.default = {}
        # switch
        self.switch = True
        self.filter = []
        # style config
        self.config_filename = config_filename if config_filename.endswith('.py') else config_filename + '.py'
        self.root_path = os.getcwd()
        self.config = Config(printme=self, root_path=self.root_path)
        self.config.from_pyfile(self.config_filename)
        # log
        self.log_output = log_output
        if log_output:
            self.queue = Queue(-1)
            t = threading.Thread(target=log_to, args=(self, log_name))
            t.start()

    def set_config(self):
        for k, v in self.config.items():
            if k not in self.level_list:
                continue
            default = self.default[k] = {}
            style = self.box[k] = {}
            for t in self.term:
                default.update({t: v[t].pop("DEFAULT", lambda: "")})
                sett = setting(**v[t])
                style.update({f'{t}0': sett[0], f'{t}1': sett[1]})

    def show(self, level, *args, **kwargs):
        style = self.box[level.upper()]
        default = self.default[level.upper()]
        data = {}
        for i in self.term:
            data[i] = kwargs.pop(i, default[i]())
        data['message'] = " ".join([str(i) for i in args])
        if self.log_output:
            msg = self.raw_template.format(**data) + "\n"
            self.queue.put(msg)
        data.update(style)
        data['message'] = " ".join([str(i) for i in args])
        print(self.template.format(**data), sep=kwargs.get('sep', " "), end=kwargs.get('end', "\n"),
              file=kwargs.get('file', None))

    @level_wrap
    def info(self, *args, **kwargs):
        self.show(INFO, *args, **kwargs)

    @level_wrap
    def debug(self, *args, **kwargs):
        self.show(DEBUG, *args, **kwargs)

    @level_wrap
    def error(self, *args, **kwargs):
        self.show(ERROR, *args, **kwargs)

    @level_wrap
    def warn(self, *args, **kwargs):
        self.show(WARN, *args, **kwargs)

    @level_wrap
    def success(self, *args, **kwargs):
        self.show(SUCCESS, *args, **kwargs)



from datetime import datetime
from colour_printing.style import setting
from colour_printing.switch import Switch

MESSAGE_STYLE = 'message_style'
FLAG_STYLE = 'flag_style'
TIME_STYLE = 'time_style'


class Pen(object):
    __get_time = lambda _: datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S.%f')[:-3]
    __flag_len = 7

    def __init__(self, flag_name: str):
        self.__flag_name = flag_name.upper()
        self.__config = {}
        self.__cal_flag_len()

    def __cal_flag_len(self):
        if Pen.__flag_len < len(self.__flag_name):
            Pen.__flag_len = len(self.__flag_name)

    def message_style(self, mode='', fore='', back=''):
        self.__config[MESSAGE_STYLE] = setting(mode=mode, fore=fore, back=back)
        return self

    def time_style(self, mode='', fore='', back=''):
        self.__config[TIME_STYLE] = setting(mode=mode, fore=fore, back=back)
        return self

    def flag_style(self, mode='', fore='', back=''):
        self.__config[FLAG_STYLE] = setting(mode=mode, fore=fore, back=back)
        return self

    def __user_setting(self):
        message_style = self.__config.get(MESSAGE_STYLE, setting())
        time_style = self.__config.get(TIME_STYLE, setting())
        flag_style = self.__config.get(FLAG_STYLE, setting())
        flag = '{0}{1}{2}'.format(flag_style[0], f"{self.__flag_name.center(self.__flag_len, ' ')}", flag_style[1])
        time = '{0}{1}{2}'.format(time_style[0], self.__get_time(), time_style[1])
        return flag, time, message_style

    def __call__(self, *args, **kwargs):
        if not Switch.signal or self.__flag_name.lower() in Switch.filter or self.__flag_name in Switch.filter:
            return
        flag, time, message_style = self.__user_setting()
        str_temp = []
        for s in args:
            str_temp.append(f'{message_style[0]}{s}{message_style[1]}')
        print(time, flag, *str_temp, **kwargs)



import errno
import os
import sys
import types


def import_string(import_name, silent=False):
    import_name = str(import_name).replace(":", ".")
    try:
        try:
            __import__(import_name)
        except ImportError:
            if "." not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit(".", 1)
        module = __import__(module_name, globals(), locals(), [obj_name])
        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e)

    except ImportError as e:
        raise e


string_types = (str,)
integer_types = (int,)

"""从.py文件导入"""


def level_template(level_name, term):
    res = ""
    res += "%s = {" % level_name
    for t in term:
        temp = """
    '%s': {
        "DEFAULT": %s,  # <-- Must be function name or lambda expression
        "fore": Fore.CYAN,
        "back": Back,
        "mode": Mode,
    },
""" % (t, t + "_default")
        res += temp
    res += "}\n\n"
    return res


def new_pyfile_template(level_list, term):
    res = """\"""
#                     *Colour-printing Reference*
#########################################################################################
#   @'fore': # 前景色         @'back':# 背景              @'mode':# 显示模式               # 
#            'black': 黑色            'black':  黑色              'normal': 终端默认设置   # 
#            'red': 红色              'red':  红色                'bold':  高亮显示        # 
#            'green': 绿色            'green': 绿色               'underline':  使用下划线 #
#            'yellow': 黄色           'yellow': 黄色              'blink': 闪烁           # 
#            'blue':  蓝色            'blue':  蓝色               'invert': 反白显示       #    
#            'purple':  紫红色        'purple':  紫红色            'hide': 不可见          #    
#            'cyan':  青蓝色          'cyan':  青蓝色                                     #
#            'white':  白色           'white':  白色                                     #
#########################################################################################
\"""
"""
    # lib
    res += """from datetime import datetime
from colour_printing import Mode, Fore, Back\n
get_time = lambda: datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S.%f')[:-3]\n
"""
    # default
    for t in term:
        res += '''%s_default = lambda: ""\n\n''' % t
    # style
    for l in level_list:
        res += level_template(l, term)
    return res


class Config(dict):
    def __init__(self, printme, root_path):
        dict.__init__(self)
        self.printme = printme
        self.root_path = root_path
        self.config_str = ""

    def create_py_file(self, filename):
        with open(filename, 'w')as f:
            f.write(new_pyfile_template(self.printme.level_list, self.printme.term))

    def from_pyfile(self, filename, silent=False):
        file_path = os.path.join(self.root_path, filename)
        if not os.path.exists(file_path):
            self.create_py_file(file_path)
            print(f'[*]Tip>> 文件不存在,现已创建path: {file_path}')
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename, mode='rb') as config_file:
                config = config_file.read()
                self.config_str = config.decode()
                exec(compile(config, filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (
                    errno.ENOENT, errno.EISDIR, errno.ENOTDIR
            ):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        if isinstance(obj, string_types):
            obj = import_string(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)
        self.printme.set_config()  # 渲染
